# DELPHI Simple Perl Chatbot
# Author: Scott Cornaby
# www.scottcornaby.com
#
# A simple chatbot built to demonstrate simple pattern matching algorithims
# along with several aspects of test driven development
#
# The current rule set results in a chatbot that acts like a magic 8 ball, giving random
# yes and no answers to simple questions. This behavior can be modified by changing the
# chatPatterns array
#
# This code can be freely used and modified for any purpose private, educational or commercial 
# with or without attribution. Seriously, do whatever you want with.

package DELPHI;

use strict;
use warnings;

my $commonAdjectives=qr/(?:this|the|that|a|an|his|hers|my|your|their|big|blue|good|new|first|last|long|great|little|own|other|old|right|big|high|different|small|large|next|early|young|important|few|public|bad|same|able)/;
my $noncaptureAdjectiveChain=qr/(?:$commonAdjectives )*/; #Zero or more adjectives with a space after each word

# The core of this rule based pattern matching chatbot.
# Each item in the chatPatterns array is a two item array
# The first item is a stored regular expression that will be compared agains the user's input
# The second item is an array of possible responses that the chatbot should choose from when it finds a match
# UIF# indicates a location where a capture group from the regular expression should be substituted into the output
my @chatPatterns;

push(@chatPatterns,
		[qr/\Ahelp\z/i,
			["Just type a question and hit enter and I'll do my best to answer it. Yes or No questions like \"Did I do well on my test?\" are best"]]);

push(@chatPatterns, 
		[qr/[a-zA-Z]+ or [a-zA-Z]+.*\?\z/,
			["Fate indicates the former",
			"I have a good feeling about the later",
			"The first option. Wait, no the second. Wait, wait, the first. I mean it this time!",
			"Let me flip a coin... the first one",
			"Let me flip a coin... the second one",
			"The later, unless it's actually the former",
			"Whatever you think is correct"]
		]);

push(@chatPatterns, 
		[qr/\ADo (.+)\?\z/i, 
			["Fate indicates that UIF0",
			"I don't think that UIF0",
			"Athena doesn't think so, and she's the Greek god of wisdom!",
			"I think it should be obvious that UIF0",
			"Of course not. Why would you even think that UIF0?"]
		]);
			
push(@chatPatterns, 
		[qr/\A(Did|Does) ($noncaptureAdjectiveChain[a-zA-Z]+) (.+)\?\z/i, 
			["Fate indicates that UIF1 UIF0 UIF2",
			"The spirits whisper \"UIF1 UIF0 not UIF2\"",
			"I don't think that UIF1 UIF0 UIF2, but let me double check my cheat sheet...",
			"Meh, UIF1 probably UIF0 UIF2, but I'm not really interested in that kind of stuff",
			"Trust your instincts on this one"]
		]);

push(@chatPatterns, 
		[qr/\A(is|are|am|was|were) ($noncaptureAdjectiveChain[a-zA-Z]+) (.+)\?\z/i, 
			["Fate indicates that UIF1 UIF0 UIF2",
			"The stars are clear: UIF1 UIF0 not UIF2",
			"UIF1 can't be UIF2. That would mess up the crossword puzzle I'm working on",
			"I think that UIF1 UIF0 UIF2, but you should probably double check with someone else",
			"Hmmm... echoes of cosmic wisdom show that UIF1 UIF0 not UIF2",
			"I see no reason why UIF1 wouldn't be UIF2"]
		]);
			
push(@chatPatterns, 
		[qr/\AWill ($noncaptureAdjectiveChain[a-zA-Z]+) (.+)\?\z/i, 
			["I predict that UIF0 will UIF1",
			"Based on these tea leaves it seems UIF0 will not UIF1",
			"UIF0 will UIF1, assuming the mole people don't invade next Tuesday",
			"I lost my tarot cards... but I have some playing cards. They predict that UIF0 will UIF1",
			"I lost my tarot cards... but I have some playing cards. They predict that UIF0 will not UIF1",
			"My advanced AI is 67% certain that UIF0 will not UIF1"]
		]);

push(@chatPatterns,
		[qr/\AWhy (.+)\?\z/i,
			["Because of reasons",
			"For important cosmic reasons",
			"You know, I'm really not sure either. Let me forward that question to my supervisor",
			"Do you really have to know why? Sometimes it is better to just accept things",
			"If you study enough advanced physics someday you might understand",
			"Why not?",
			"Because of a decision made a long time ago by someone you've never met",
			"Because you subconsciously believe that's how the universe works"]
		]);
		
push(@chatPatterns,
		[qr/\AWhat.*questions/i,
			["Yes or no questions like \"Will I have good luck tomorrow?\" are the easiest for me to answer"]]);
			
push(@chatPatterns,
		[qr/\ACan.*\bI\b/i,
			["The only thing I can really do is answer simple questions like \"Will it be sunny tomorrow?\""]]);
			
push(@chatPatterns, 
		[qr/\ACan ($noncaptureAdjectiveChain[a-zA-Z]+) (.+)\?\z/i, 
			["Fate indicates that UIF0 can UIF1",
			"My \"Big Book O' Wisdom\" says that UIF0 can't UIF1"]
		]);

push(@chatPatterns, 
		[qr/\?/,
			["I'm sorry, could you try rewording that as a yes or no question?",
			"I saw a question mark, but I didn't understand the question. I'm very sorry, but you could try again with a yes or no question?",
			"That looked like a question, but I'm not sure how to respond. Maybe you could simplify it to a yes or no question",
			"I'm sorry, could you think of a simpler way to ask that question? Maybe as a yes or no question?"]
		]);

push(@chatPatterns, 
		[qr/\A(Why|Is|Are|Do|Does|Will)/i,
			["Did you forget a question mark? Grammar is important!",
			"If you're asking a question, remember to use a question mark. I'm very picky",
			"That looked like a question, but without a question mark I'm really not sure",
			"I think that was a question, but I'd know for sure if you used a question mark"]
		]);

push(@chatPatterns,
		[qr/.*/,
			["I don't want to talk about that. Please feel free to ask me why",
			"I'm confused. Try a simple yes or no question instead",
			"I'm really good at yes no questions. Try one of those",
			"I'm sorry, I'm not sure what to say about that. Try a question like 'Why is the sky blue?'"]
		]);

# Uses the rules in the chatPattern array to produce an appropriate response to user input
# Arguments: string user input, optional boolean random (choose whether response should be reandomized or not)
# Returns: string
sub generateResponse{
	my $userInput = $_[0];
	my $beRandom = $_[1];
	$userInput = switchFirstAndSecondPerson($userInput);

	# Loop through all rules. When a match is found for the user's input generate and return a response
	foreach my $chatPattern (@chatPatterns){
		
		if(my @UIF = ($userInput =~ $chatPattern->[0])){
			my $response;
			if($beRandom){
				my $numberOfResponses = scalar(@{ $chatPattern->[1] });
				$response = $chatPattern->[1][rand $numberOfResponses];
			}
			else{
				$response = $chatPattern->[1][0];
			}
			# Substitute pieces of the users input into chatbot output in all locations where UIF# pattern is found
			for(my $i=0; $i<@UIF; $i++){
				my $find = "UIF$i";
				my $replace = $UIF[$i];
				$response =~ s/$find/$replace/g;
			}
			return uc($response);
		}
	}
	# chatPatterns should always contain a base rule that matches everything not caught by higher priority rules
	# Exiting the foreach loop without finding a rule suggests a mistake in designing the rule set
	return "Base Case Failure Error!";
}

#Dictionaries used to help the switchFirstAndSecondPerson function do its job
my @wordsToPlaceholders;

$wordsToPlaceholders[0][0]=qr/\bI\b/i;
$wordsToPlaceholders[0][1]='DELPHIyou';

$wordsToPlaceholders[1][0]=qr/\bme\b/i;
$wordsToPlaceholders[1][1]='DELPHIyou';

$wordsToPlaceholders[2][0]=qr/\bmine\b/i;
$wordsToPlaceholders[2][1]='DELPHIyours';

$wordsToPlaceholders[3][0]=qr/\bmy\b/i;
$wordsToPlaceholders[3][1]='DELPHIyour';

$wordsToPlaceholders[4][0]=qr/\byou\b/i;
$wordsToPlaceholders[4][1]='DELPHIi';

$wordsToPlaceholders[5][0]=qr/\byour\b/i;
$wordsToPlaceholders[5][1]='DELPHImy';

$wordsToPlaceholders[6][0]=qr/\byours\b/i;
$wordsToPlaceholders[6][1]='DELPHImine';

my @placeholdersToWords;

$placeholdersToWords[0][0]=qr/DELPHIyou/;
$placeholdersToWords[0][1]='you';

$placeholdersToWords[1][0]=qr/DELPHIyour/;
$placeholdersToWords[1][1]='your';

$placeholdersToWords[2][0]=qr/DELPHIyours/;
$placeholdersToWords[2][1]='yours';

$placeholdersToWords[3][0]=qr/DELPHIi/;
$placeholdersToWords[3][1]='I';

$placeholdersToWords[4][0]=qr/DELPHImine/;
$placeholdersToWords[4][1]='mine';

$placeholdersToWords[5][0]=qr/DELPHImy/;
$placeholdersToWords[5][1]='my';

# Switch around first and second person words in a string
# Ex: You want to talk to me -> I want to talk to you
# Argument: string input
# Returns: string
sub switchFirstAndSecondPerson{
	my $input =$_[0];
	
	foreach my $wordToPlaceholder (@wordsToPlaceholders){
		$input =~ s/$wordToPlaceholder->[0]/$wordToPlaceholder->[1]/g;
	}
	
	foreach my $placeholderToWord (@placeholdersToWords){
		$input =~ s/$placeholderToWord->[0]/$placeholderToWord->[1]/g;
	}
	
	return $input;
}

1;
